#include <mailutils/mailutils.h>
#include <mailutils/imaputil.h>
#include <mailutils/nls.h>
#include <magic.h>

#define PACKAGE_BUG_ADDRESS "gray@gnu.org"
#define PACKAGE_URL "https://puszcza.gnu.org.ua/projects/imapbin"

#undef EXIT_SUCCESS
#undef EXIT_FAILURE

enum
  {
    EXIT_SUCCESS,
    EXIT_FAILURE,
    EXIT_USAGE
  };

mu_header_t add_header;
magic_t mag;
char *default_mime_type = "application/octet-stream";
int dry_run;
char *magic_files;

static void
cli_header_helper (struct mu_parseopt *po, char const *name, char const *value)
{
  int rc;

  if (!add_header)
    {
      if ((rc = mu_header_create (&add_header, NULL, 0)) != 0)
	{
	  mu_parseopt_error (po, "can't create header: %s", mu_strerror (rc));
	  exit (po->po_exit_error);
	}
    }

  if ((rc = mu_header_append (add_header, name, value)) != 0)
    {
      mu_parseopt_error (po, "can't append header: %s", mu_strerror (rc));
      exit (po->po_exit_error);
    }
}

static void
cli_append (struct mu_parseopt *po, struct mu_option *opt, char const *arg)
{
  size_t len;
  char *name, *value;

  len = strcspn (arg, ":");
  if (arg[len] == 0)
    {
      mu_parseopt_error (po, "invalid header: %s", arg);
      exit (po->po_exit_error);
    }
  name = mu_alloc (len + 1);
  memcpy (name, arg, len);
  name[len] = 0;
  value = mu_strdup (mu_str_skip_class (arg + len + 1, MU_CTYPE_SPACE));
  cli_header_helper (po, name, value);
  free (name);
  free (value);
}

static void
cli_from (struct mu_parseopt *po, struct mu_option *opt, char const *arg)
{
  cli_header_helper (po, MU_HEADER_FROM, arg);
}

static void
cli_to (struct mu_parseopt *po, struct mu_option *opt, char const *arg)
{
  cli_header_helper (po, MU_HEADER_TO, arg);
}

static void
cli_subject (struct mu_parseopt *po, struct mu_option *opt, char const *arg)
{
  cli_header_helper (po, MU_HEADER_SUBJECT, arg);
}

static char const *
basename (char const *name)
{
  char const *p = strrchr (name, '/');
  return p ? p + 1 : name;
}

static char const *text_types[] = {
  "text/*",
  "application/*shell",
  "application/shellscript",
  "*/x-csrc",
  "*/x-csource",
  "*/x-diff",
  "*/x-patch",
  "*/x-perl",
  "*/x-php",
  "*/x-python",
  "*/x-sh",
  NULL
};

static int
is_text_type (char const *mime_type)
{
  int i;
  for (i = 0; text_types[i]; i++)
    {
      if (mu_imap_wildmatch_ci (text_types[i], mime_type, 0) == 0)
	return 1;
    }
  return 0;
}

static void
mime_attach_file (mu_mime_t mime, char const *name)
{
  int rc;
  mu_stream_t str;
  mu_message_t part;
  char const *b;
  char *mime_type = NULL;
  char *encoding = "base64";

  rc = mu_file_stream_create (&str, name, MU_STREAM_READ);
  if (rc)
    {
      mu_error ("can't open file %s: %s", name, mu_strerror (rc));
      exit (EXIT_FAILURE);
    }

  if (mag)
    {
      char const *descr = magic_file (mag, name);

      if (descr)
	{
	  size_t len = strcspn (descr, ";");
	  mime_type = mu_alloc (len + 1);
	  memcpy (mime_type, descr, len);
	  mime_type[len] = 0;
	}
    }

  if (mime_type && is_text_type (mime_type))
    encoding = "quoted-printable";

  b = basename (name);
  rc = mu_attachment_create (&part,
			     mime_type ? mime_type : default_mime_type,
			     encoding, b, b);
  free (mime_type);
  if (rc)
    {
      mu_error ("%s: can't create attachment: %s", name, mu_strerror (rc));
      exit (EXIT_FAILURE);
    }

  rc = mu_attachment_copy_from_stream (part, str);
  if (rc)
    {
      mu_error ("%s: can't copy to attachment: %s", name, mu_strerror (rc));
      exit (EXIT_FAILURE);
    }

  mu_stream_destroy (&str);

  rc = mu_mime_add_part (mime, part);
  if (rc)
    {
      mu_error ("%s: can't add mime part: %s", name, mu_strerror (rc));
      exit (EXIT_FAILURE);
    }

  mu_message_unref (part);
}

#define MU_DATETIME_RFC822_LENGTH 31

void
add_hdr_date (mu_header_t hdr)
{
  struct timeval tv;
  struct tm *tm;
  char buf[MU_DATETIME_RFC822_LENGTH+1];
  int rc;

  gettimeofday (&tv, NULL);
  tm = localtime (&tv.tv_sec);
  mu_strftime (buf, sizeof (buf), MU_DATETIME_FORM_RFC822, tm);
  rc = mu_header_set_value (hdr, MU_HEADER_DATE, buf, 1);
  if (rc)
    mu_diag_funcall (MU_DIAG_ERROR, "mu_header_set_value", MU_HEADER_DATE, rc);
}

void
add_hdr_from (mu_header_t hdr, mu_message_t msg)
{
  int rc;
  mu_envelope_t env;
  char const *sender;

  mu_message_get_envelope (msg, &env);
  mu_envelope_sget_sender (env, &sender);
  rc = mu_header_set_value (hdr, MU_HEADER_FROM, sender, 1);
  if (rc)
    mu_diag_funcall (MU_DIAG_ERROR, "mu_header_set_value", MU_HEADER_FROM, rc);
}

void
add_hdr_to (mu_header_t hdr)
{
  int rc = mu_header_set_value (hdr, MU_HEADER_TO, "", 1);
  if (rc)
    mu_diag_funcall (MU_DIAG_ERROR, "mu_header_set_value", MU_HEADER_TO, rc);
}

void
add_hdr_subject (mu_header_t hdr, int argc, char **argv)
{
  int rc;
  char *buf = NULL;

  if (argc == 1)
    mu_asprintf (&buf, "Upload of %s", basename (argv[0]));
  else if (argc == 2)
    mu_asprintf (&buf, "Upload of %s and %s", basename (argv[0]), basename (argv[1]));
  else
    mu_asprintf (&buf, "Upload of %s and %d other files", basename (argv[0]),
		 argc - 1);
  rc = mu_header_set_value (hdr, MU_HEADER_SUBJECT, buf, 1);
  if (rc)
    mu_diag_funcall (MU_DIAG_ERROR, "mu_header_set_value", MU_HEADER_SUBJECT, rc);
  free (buf);
}


struct mu_option options[] = {
  { "append",   'a', "HEADER:VALUE", MU_OPTION_DEFAULT,
    "append given header to the message being constructed",
    mu_c_string, NULL, cli_append },
  { "from",     'F', "EMAIL", MU_OPTION_DEFAULT,
    "set from address",
    mu_c_string, NULL, cli_from },
  { "to",       't', "EMAIL", MU_OPTION_DEFAULT,
    "set recipient address",
    mu_c_string, NULL, cli_to },
  { "subject",  's', "TEXT", MU_OPTION_DEFAULT,
    "set subject",
    mu_c_string, NULL, cli_subject },
  { "dry-run",  'n', NULL, MU_OPTION_DEFAULT,
    "don't actually submit the message",
    mu_c_bool, &dry_run },
  { "magic",    'm', "MAGICFILES", MU_OPTION_DEFAULT,
    "alternate list of files and directories containing file magic database",
    mu_c_string, &magic_files },
  MU_OPTION_END
};

static char *capa[] = {
  "debug",
  "mailbox",
  NULL
};

void
version_hook (struct mu_parseopt *po, mu_stream_t stream)
{
	mu_stream_printf (stream, "%s (%s) %s\n", mu_program_name,
			  PACKAGE_NAME, PACKAGE_VERSION);
	mu_stream_printf (stream, "Copyright (C) 2021, 2025 Sergey Poznyakoff");
	mu_stream_printf (stream, "\
\n\
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\nThis is free software: you are free to change and redistribute it.\n\
There is NO WARRANTY, to the extent permitted by law.\n\
\n\
");
}

int
main (int argc, char *argv[])
{
  int rc, i;
  mu_mailbox_t mbox;
  mu_message_t msg;
  mu_mime_t mime;
  mu_iterator_t itr;
  mu_header_t outhdr;

  mu_registrar_record (mu_imap_record);
  mu_registrar_record (mu_imaps_record);

  mu_set_program_name (argv[0]);
  mu_cli_simple (argc, argv,
		 MU_CLI_OPTION_OPTIONS, options,
		 MU_CLI_OPTION_PROG_DOC, "packs FILEs into a MIME message and appends it to IMAP mailbox at URL",
		 MU_CLI_OPTION_PROG_ARGS, "URL FILE [FILE...]",
		 MU_CLI_OPTION_PACKAGE_NAME, PACKAGE_NAME,
		 MU_CLI_OPTION_BUG_ADDRESS, PACKAGE_BUG_ADDRESS,
		 MU_CLI_OPTION_PACKAGE_URL, PACKAGE_URL,
		 MU_CLI_OPTION_VERSION_HOOK, version_hook,
		 MU_CLI_OPTION_RETURN_ARGC, &argc,
		 MU_CLI_OPTION_RETURN_ARGV, &argv,
		 MU_CLI_OPTION_CAPABILITIES, capa,
		 MU_CLI_OPTION_EX_USAGE, EXIT_USAGE,
		 MU_CLI_OPTION_END);

  if (argc < 2)
    {
      mu_error ("not enough arguments");
      exit (EXIT_USAGE);
    }

  rc = mu_mime_create (&mime, NULL, MU_MIME_MULTIPART_MIXED);
  if (rc)
    {
      mu_error ("can't create MIME: %s", mu_strerror (rc));
      exit (EXIT_FAILURE);
    }

  mag = magic_open (MAGIC_MIME|MAGIC_SYMLINK);
  if (!mag)
    {
      mu_error ("can't open file magic database: %s", mu_strerror (errno));
    }
  magic_load (mag, magic_files);

  for (i = 1; i < argc; i++)
    {
      mime_attach_file (mime, argv[i]);
    }

  rc = mu_mime_get_message (mime, &msg);
  if (rc)
    {
      mu_error ("can't create message: %s", mu_strerror (rc));
      exit (EXIT_FAILURE);
    }

  /* Copy rest of headers from the original message */
  rc = mu_message_get_header (msg, &outhdr);
  if (rc)
    {
      mu_diag_funcall (MU_DIAG_ERROR, "mu_message_get_header", NULL, rc);
      exit (EXIT_FAILURE);
    }

  add_hdr_date (outhdr);
  add_hdr_from (outhdr, msg);
  add_hdr_to (outhdr);
  add_hdr_subject (outhdr, argc - 1, argv + 1);

  if (add_header)
    {
      rc = mu_header_get_iterator (add_header, &itr);
      if (rc)
	{
	  mu_diag_funcall (MU_DIAG_ERROR, "mu_header_get_iterator", NULL, rc);
	  exit (EXIT_FAILURE);
	}
      for (mu_iterator_first (itr); !mu_iterator_is_done (itr);
	   mu_iterator_next (itr))
	{
	  const char *name, *value;

	  if (mu_iterator_current_kv (itr, (const void **)&name,
				      (void**)&value) == 0)
	    {
	      if (mu_c_strcasecmp (name, MU_HEADER_RECEIVED) == 0
		  || mu_c_strncasecmp (name, "X-", 2) == 0)
		mu_header_append (outhdr, name, value);
	      else if (mu_c_strcasecmp (name, MU_HEADER_MIME_VERSION) == 0 ||
		       mu_c_strncasecmp (name, "Content-", 8) == 0)
		{
		  mu_error ("%s: not setting header", name);
		  continue;
		}
	      else
		mu_header_set_value (outhdr, name, value, 1);
	    }
	}
      mu_iterator_destroy (&itr);
    }

  if (dry_run)
    {
      mu_stream_t str;

      mu_message_get_streamref (msg, &str);
      mu_stream_copy (mu_strout, str, 0, NULL);
      mu_stream_destroy (&str);
    }
  else
    {
      if ((rc = mu_mailbox_create_default (&mbox, argv[0])) != 0)
	{
	  mu_error ("can't open mailbox %s: %s", argv[0],
		    mu_strerror (rc));
	  exit (EXIT_FAILURE);
	}
      if ((rc = mu_mailbox_open (mbox, MU_STREAM_CREAT|MU_STREAM_APPEND)) != 0)
	{
	  mu_error ("can't open mailbox %s: %s", argv[0], mu_strerror (rc));
	  exit (EXIT_FAILURE);
	}

      rc = mu_mailbox_append_message (mbox, msg);
      if (rc)
	mu_error ("can't append message: %s", mu_strerror (rc));

      mu_mailbox_close (mbox);
      mu_mailbox_destroy (&mbox);
    }

  mu_mime_destroy (&mime);

  if (rc)
    return EXIT_FAILURE;
  return EXIT_SUCCESS;
}
